# OpenMP support for GDB, the GNU debugger
#
# Copyright (C) 2015 Free Software Foundation, Inc.
#
# Contributed by Intel Corporation, <walfred.tedeschi@intel.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import gdb
import openmp_rtl_support
import openmp_commons

def print_items(collection):
    for item in sorted(collection):
        if collection[item] is not None:
            print collection[item]

def print_sparse_items(collection, desired_item):
    already_printed = False
    for item in sorted(collection):
        if collection[item] is not None and item in desired_item:
            print collection[item]
            already_printed = True
    if not already_printed:
        print "Item(s) not found."

class AbstractOpenMPCommand(gdb.Command):

    def __init__(self):
        super(AbstractOpenMPCommand, self).__init__(self._omp_cmd_syntax, gdb.COMMAND_DATA)

    def setup_header(self):
        pass

    def invoke(self, arg, from_tty):
        try:
            rtl_interface = openmp_rtl_support.get_openmp_interface()
        except openmp_commons.OMPException as e:
            print e
            return
        items = getattr(rtl_interface, self._omp_invoke_info)
        if not items:
            print 'No %s found.' % self._omp_invoke_info
            return
        if not arg:
            self.setup_header()
            print_items(items)
        else:
            desired_items = []
            try:
                desired_items = [int(i) for i in arg.strip().split()]
            except:
                print "Expecting integers as arguments."
            if desired_items:
                self.setup_header()
                print_sparse_items(items, desired_items)


class OpenMPBarrierInfo(AbstractOpenMPCommand):
    '''Display the barriers found on a running OpenMP application'''
    _omp_cmd_syntax = 'info omp-barrier'
    _omp_invoke_info = 'barriers'


class OpenMPLockInfo(AbstractOpenMPCommand):
    '''Display OpenMP locks information'''
    _omp_cmd_syntax = 'info omp-lock'
    _omp_invoke_info = 'locks'

    def setup_header(self):
         openmp_commons.Lock.print_header()


class OpenMPTaskInfo(AbstractOpenMPCommand):
    '''Display information about tasks running on OpenMP.'''
    _omp_cmd_syntax = 'info omp-task'
    _omp_invoke_info = 'tasks'


class OpenMPTaskWaitInfo(AbstractOpenMPCommand):
    '''Display taskwaits found on  an OpenMP application'''
    _omp_cmd_syntax = 'info omp-taskwait'
    _omp_invoke_info = 'taskwaits'


class OpenMPTeamInfo(AbstractOpenMPCommand):
    '''Display '''
    _omp_cmd_syntax = 'info omp-team'
    _omp_invoke_info = 'team'


class OpenMPTaskdep(AbstractOpenMPCommand):
    '''Display dependent tasks in the order that they will be processed'''
    _omp_cmd_syntax = 'info omp-dependency'
    _omp_invoke_info = 'dependency'

    def setup_header(self):
        openmp_commons.TaskDependency.print_header()


OpenMPTaskInfo()
OpenMPLockInfo()
OpenMPTeamInfo()
OpenMPTaskWaitInfo()
OpenMPBarrierInfo()
OpenMPTaskdep()
