# OpenMP support for GDB, the GNU debugger
#
# Copyright(C) 2015 Free Software Foundation, Inc.
#
# Contributed by Intel Corporation, <walfred.tedeschi@intel.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import ctypes
import os

import gdb

from openmp_commons import *

c_uint32 = ctypes.c_uint32

debug_symbol = '__kmp_omp_debug_struct_info'
sizeof_debug_str = '%s.%s' % (debug_symbol, 'sizeof_this_structure')
version_debug_str = '%s.%s' % (debug_symbol, 'lib_version')

# globals...
threads_debug_str = '%s.%s' % (debug_symbol, 'threads')
roots_debug_str = '%s.%s' % (debug_symbol, 'roots')
capacity_debug_str = '%s.%s' % (debug_symbol, 'capacity')
monitor_debug_str = '%s.%s' % (debug_symbol, 'monitor')
lock_table_str = '%s.%s' % (debug_symbol, 'lock_table')
func_microtask_str = '%s.%s' % (debug_symbol, 'func_microtask')
func_fork_str = '%s.%s' % (debug_symbol, 'func_fork')
func_fork_teams_str = '%s.%s' % (debug_symbol, 'func_fork_teams')
team_counter_str = '%s.%s' % (debug_symbol, 'team_counter')
task_counter_str = '%s.%s' % (debug_symbol, 'task_counter')
nthr_info_str = '%s.%s' % (debug_symbol, 'nthr_info')
address_width_str = '%s.%s' % (debug_symbol, 'address_width')
indexed_locks_str = '%s.%s' % (debug_symbol, 'indexed_locks')
last_barrier_str = '%s.%s' % (debug_symbol, 'last_barrier')
deque_size_str = '%s.%s' % (debug_symbol, 'deque_size')

# thread structure information.
th_sizeof_struct_str = '%s.%s' % (debug_symbol, 'th_sizeof_struct')
th_info_str = '%s.%s' % (debug_symbol, 'th_info')
th_team_str = '%s.%s' % (debug_symbol, 'th_team')
th_root_str = '%s.%s' % (debug_symbol, 'th_root')
th_serial_team_str = '%s.%s' % (debug_symbol, 'th_serial_team')
th_ident_str = '%s.%s' % (debug_symbol, 'th_ident')
th_spin_here_str = '%s.%s' % (debug_symbol, 'th_spin_here')
th_next_waiting_str = '%s.%s' % (debug_symbol, 'th_next_waiting')
th_task_team_str = '%s.%s' % (debug_symbol, 'th_task_team')
th_current_task_str = '%s.%s' % (debug_symbol, 'th_current_task')
th_task_state_str = '%s.%s' % (debug_symbol, 'th_task_state')
th_bar_str = '%s.%s' % (debug_symbol, 'th_bar')
th_b_worker_arrived_str = '%s.%s' % (debug_symbol, 'th_b_worker_arrived')

# entry address for teams construct
th_teams_microtask_str = '%s.%s' % (debug_symbol, 'th_teams_microtask')
# initial level of teams construct
th_teams_level_str = '%s.%s' % (debug_symbol, 'th_teams_level')
# number of teams in a league
th_teams_nteams_str = '%s.%s' % (debug_symbol, 'th_teams_nteams')
# number of threads in each team of the league
th_teams_nth_str = '%s.%s' % (debug_symbol, 'th_teams_nth')

# kmp_desc structure(for info field above)
ds_sizeof_struct_str = '%s.%s' % (debug_symbol, 'ds_sizeof_struct')
ds_tid_str = '%s.%s' % (debug_symbol, 'ds_tid')  # team thread id
ds_gtid_str = '%s.%s' % (debug_symbol, 'ds_gtid')  # global thread id
ds_thread_str = '%s.%s' % (debug_symbol, 'ds_thread')  # native thread id

# team structure information
t_sizeof_struct_str = '%s.%s' % (debug_symbol, 't_sizeof_struct')
# tid of master in parent team
t_master_tid_str = '%s.%s' % (debug_symbol, 't_master_tid')
# location of parallel region
t_ident_str = '%s.%s' % (debug_symbol, 't_ident')
t_parent_str = '%s.%s' % (debug_symbol, 't_parent')  # parent team
t_nproc_str = '%s.%s' % (debug_symbol, 't_nproc')  # # team threads
t_threads_str = '%s.%s' % (debug_symbol, 't_threads')  # array of threads
# # levels of serialized teams
t_serialized_str = '%s.%s' % (debug_symbol, 't_serialized')
t_id_str = '%s.%s' % (debug_symbol, 't_id')  # unique team id
t_pkfn_str = '%s.%s' % (debug_symbol, 't_pkfn')
t_task_team_str = '%s.%s' % (
    debug_symbol, 't_task_team')  # task team structure
# taskdata for the thread's implicit task
t_implicit_task_str = '%s.%s' % (debug_symbol, 't_implicit_task')
t_cancel_request_str = '%s.%s' % (debug_symbol, 't_cancel_request')
t_bar_str = '%s.%s' % (debug_symbol, 't_bar')
# increased by 1 when master arrives to a barrier
t_b_master_arrived_str = '%s.%s' % (debug_symbol, 't_b_master_arrived')
# increased by one when all the threads arrived
t_b_team_arrived_str = '%s.%s' % (debug_symbol, 't_b_team_arrived')

# root structure information
r_sizeof_struct_str = '%s.%s' % (debug_symbol, 'r_sizeof_struct')
r_root_team_str = '%s.%s' % (debug_symbol, 'r_root_team')  # team at root
# hot team for this root
r_hot_team_str = '%s.%s' % (debug_symbol, 'r_hot_team')
r_uber_thread_str = '%s.%s' % (debug_symbol, 'r_uber_thread')  # root thread
# unique root id(if available)
r_root_id_str = '%s.%s' % (debug_symbol, 'r_root_id')

# ident structure informatio
id_sizeof_struct_str = '%s.%s' % (debug_symbol, 'id_sizeof_struct')
# address of string '_str = '%s.%s'   %(debug_symbol, '')file_str =
# '%s.%s'   %(debug_symbol, '')func_str = '%s.%s'   %(debug_symbol,
# '')line1_str = '%s.%s'   %(debug_symbol, '')line2_str = '%s.%s'
# %(debug_symbol, '')_str = '%s.%s'   %(debug_symbol, '')'.
id_psource_str = '%s.%s' % (debug_symbol, 'id_psource')
id_flags_str = '%s.%s' % (debug_symbol, 'id_flags')

# lock structure information
lk_sizeof_struct_str = '%s.%s' % (debug_symbol, 'lk_sizeof_struct')
lk_initialized_str = '%s.%s' % (debug_symbol, 'lk_initialized')
lk_location_str = '%s.%s' % (debug_symbol, 'lk_location')
lk_tail_id_str = '%s.%s' % (debug_symbol, 'lk_tail_id')
lk_head_id_str = '%s.%s' % (debug_symbol, 'lk_head_id')
lk_next_ticket_str = '%s.%s' % (debug_symbol, 'lk_next_ticket')
lk_now_serving_str = '%s.%s' % (debug_symbol, 'lk_now_serving')
lk_owner_id_str = '%s.%s' % (debug_symbol, 'lk_owner_id')
lk_depth_locked_str = '%s.%s' % (debug_symbol, 'lk_depth_locked')
lk_lock_flags_str = '%s.%s' % (debug_symbol, 'lk_lock_flags')

# lock_table_t
# Size and layout of kmp_lock_table_t.
lt_size_of_struct_str = '%s.%s' % (debug_symbol, 'lt_size_of_struct')
lt_used_str = '%s.%s' % (debug_symbol, 'lt_used')
lt_allocated_str = '%s.%s' % (debug_symbol, 'lt_allocated')
lt_table_str = '%s.%s' % (debug_symbol, 'lt_table')

# task_team_t
tt_sizeof_struct_str = '%s.%s' % (debug_symbol, 'tt_sizeof_struct')
tt_threads_data_str = '%s.%s' % (debug_symbol, 'tt_threads_data')
tt_found_tasks_str = '%s.%s' % (debug_symbol, 'tt_found_tasks')
tt_nproc_str = '%s.%s' % (debug_symbol, 'tt_nproc')
tt_unfinished_threads_str = '%s.%s' % (debug_symbol, 'tt_unfinished_threads')
tt_active_str = '%s.%s' % (debug_symbol, 'tt_active')
# tt_state_str = '%s.%s' %(debug_symbol, 'tt_state')

# kmp_taskdata_t
td_sizeof_struct_str = '%s.%s' % (debug_symbol, 'td_sizeof_struct')
td_task_id_str = '%s.%s' % (debug_symbol, 'td_task_id')  # task id
td_flags_str = '%s.%s' % (debug_symbol, 'td_flags')  # task flags
td_team_str = '%s.%s' % (debug_symbol, 'td_team')  # team for this task
td_parent_str = '%s.%s' % (debug_symbol, 'td_parent')  # parent task
td_level_str = '%s.%s' % (debug_symbol, 'td_level')  # task testing level
td_ident_str = '%s.%s' % (debug_symbol, 'td_ident')  # task identifier
# child tasks(+ current task) not yet deallocated
td_allocated_child_tasks_str = '%s.%s' % (
    debug_symbol, 'td_allocated_child_tasks')
td_incomplete_child_tasks_str = '%s.%s' % (
    debug_symbol, 'td_incomplete_child_tasks')  # child tasks not yet complete

# Taskwait
td_taskwait_ident_str = '%s.%s' % (debug_symbol, 'td_taskwait_ident')
td_taskwait_counter_str = '%s.%s' % (debug_symbol, 'td_taskwait_counter')
# gtid + 1 of thread encountered taskwait
td_taskwait_thread_str = '%s.%s' % (debug_symbol, 'td_taskwait_thread')

# Taskgroup
# pointer to the current taskgroup
td_taskgroup_str = '%s.%s' % (debug_symbol, 'td_taskgroup')
# number of allocated and not yet complete tasks
td_task_count_str = '%s.%s' % (debug_symbol, 'td_task_count')
# request for cancellation of this taskgroup
td_cancel_str = '%s.%s' % (debug_symbol, 'td_cancel')

# Task dependency
# pointer to graph node if the task has dependencies
td_depnode_str = '%s.%s' % (debug_symbol, 'td_depnode')
# dn_successors_str = '%s.%s' %(debug_symbol, 'dn_successors')
# dn_node_str = '%s.%s' %(debug_symbol, 'dn_node')
# dn_next_str = '%s.%s' %(debug_symbol, 'dn_next')

dn_successors_str = '%s.%s' % (debug_symbol, 'dn_next')
dn_node_str = '%s.%s' % (debug_symbol, 'dn_successors')
dn_next_str = '%s.%s' % (debug_symbol, 'dn_node')

dn_task_str = '%s.%s' % (debug_symbol, 'dn_task')
dn_routine_str = '%s.%s' % (debug_symbol, 'dn_routine')
dn_npredecessors_str = '%s.%s' % (debug_symbol, 'dn_npredecessors')
dn_nrefs_str = '%s.%s' % (debug_symbol, 'dn_nrefs')

# kmp_thread_data_t
hd_sizeof_struct_str = '%s.%s' % (debug_symbol, 'hd_sizeof_struct')
hd_deque_str = '%s.%s' % (debug_symbol, 'hd_deque')
hd_deque_head_str = '%s.%s' % (debug_symbol, 'hd_deque_head')
hd_deque_tail_str = '%s.%s' % (debug_symbol, 'hd_deque_tail')
hd_deque_ntasks_str = '%s.%s' % (debug_symbol, 'hd_deque_ntasks')
hd_deque_last_stolen_str = '%s.%s' % (debug_symbol, 'hd_deque_last_stolen')


class OffsetSize(object):

    def __init__(self, offset, size):
        self.offset = offset
        self.size = size

    def __str__(self):
        return 'offset = 0x%x, size = 0x%x' % (self.offset, self.size)


class AddressSizePadding(object):

    def __init__(self, address, size, padding):
        self.addr = address
        self.size = size
        self.padding = padding

    def __str__(self):
        return 'addr = 0x%x, size = 0x%x, padding = 0x%x' % (self.addr, self.size, self.padding)


def values_for(base_address, *fields_as_strings):
    return [get_value_for(base_address, par) for par in fields_as_strings]

offsets = {}


def offset_size(element):
    if not element in offsets:
        offset = long(gdb.parse_and_eval('%s.%s' % (element, 'offset')))
        size = long(gdb.parse_and_eval('%s.%s' % (element, 'size')))
        offsets[element] = OffsetSize(offset, size)
    return offsets[element]


def get_integer_from(element):
    return long(gdb.parse_and_eval('%s' % element))


def address_size_padding(element):
    address = long(gdb.parse_and_eval('%s.%s' % (element, 'addr')))
    size = long(gdb.parse_and_eval('%s.%s' % (element, 'size')))
    padding = long(gdb.parse_and_eval('%s.%s' % (element, 'padding')))
    return AddressSizePadding(address, size, padding)


def read_integer(address, size, offset=0):
    if size != 4 and size != 8:
        raise OMPNotSupportedSize
    try:
        buffer = gdb.selected_inferior().read_memory(address + offset, size)
        data_format = '@Q' if size == 8 else '@i'
        return struct.unpack(data_format, buffer)[0]
    except:
        pass


def get_source_from_ident(address):
    try:
        source = gdb.parse_and_eval('(char*) 0x%x' % address)
        split_list = str(source).split(';')
        if len(split_list) < 4:
            return ''
        source_file = os.path.basename(split_list[1])
        function = split_list[2]
        line = split_list[3]
        return '%s %s:%s' % (function, source_file, line)
    except:
        return ''


def get_thread_handler(num):
    try:
        retval = gdb.execute('info thread %d' % num, False, True)
        retval = retval.split('\n')[1][1:]
        retval = retval.split()[2]
        return long(retval, 16)
    except:
        return 0


def get_value_for(address, string):
    val = offset_size(string)
    return read_integer(address, val.size, val.offset)


def get_address_and_size_for(address, string):
    val = offset_size(string)
    return OffsetSize(read_integer(address, val.size, val.offset), val.size)


class ArrayIterator(object):

    def __init__(self, address, structure_size, capacity):
        self.address = address
        self.current_address = address
        self.struct_size = structure_size
        self.capacity = capacity
        self.index = 0

    def advance(self):
        if self.index > self.capacity:
            return None
        self.current_address = self.address + self.struct_size * self.index
        self.index += 1
        return self.current_address

barrier_index = {'implicit': 0, 'explicit': 1, 'reduction': 2}


def get_thread_has_arrived(address, kind):
    barrier = offset_size(th_bar_str)
    index = barrier_index[kind]
    address_base = address + barrier.size * index
    thread_has_arrived, team_has_arrived = values_for(
        address_base, th_b_worker_arrived_str, th_b_worker_arrived_str)
    return values_for(address_base, th_b_worker_arrived_str)[0]


def check_size_of_debug_structure():
    sizeof = 'sizeof(%s)' % debug_symbol
    size_from_debugger = gdb.parse_and_eval(sizeof)
    size_from_rtl = gdb.parse_and_eval(sizeof_debug_str)
    return size_from_debugger == size_from_rtl


def get_ident(address):
    psource = get_value_for(address, id_psource_str)
    if psource:
        return get_source_from_ident(psource)


def get_root_team():
    root = address_size_padding(roots_debug_str)
    root_address = read_integer(root.addr, root.size, 0)
    root_address = read_integer(root_address, root.size, 0)
    root_team_address = get_value_for(root_address, r_root_team_str)
    return get_team_info(root_team_address)


def get_team_info(address):
    return TeamInternal(address)


def get_task_info(address):
    return TaskInternal(address)


def get_taskwait_info(address):
    ident, counter, thread = values_for(address, td_taskwait_ident_str,
                                        td_taskwait_counter_str,
                                        td_taskwait_thread_str)
    if ident:
        ident = get_ident(ident)
    return ident, counter, thread


def get_taskgroup_info(add):
    taskgroup = get_value_for(add, td_taskgroup_str)
    if not taskgroup:
        return
    address = taskgroup
    size_to_read = offset_size(td_taskgroup_str).size
    address = read_integer(address, size_to_read)
    td_task_count = get_value_for(add, td_task_count_str)
    td_cancel = get_value_for(add, td_cancel_str)
    try:
        print '0x%x, 0x%x, 0x%x' % (td_taskgroup, td_task_count, td_cancel)
    except:
        pass


def get_task_routine(task):
    try:
        routine = get_value_for(task, dn_routine_str)
        symbol_and_line = gdb.find_pc_line(routine)
        function = gdb.block_for_pc(routine).function
        compilation_file = symbol_and_line.symtab.filename
        line = symbol_and_line.line
        return Routine(routine, function, compilation_file, line)
    except:
        pass


# that is the successors
#
# type = struct kmp_depnode_list {
#    kmp_depnode_t *node;
#    kmp_depnode_list_t *next;
# }

def process_successor(node, address):
    future_node, next = values_for(address, dn_node_str, dn_next_str)
    if future_node:
        process_node(node, future_node)
    if next:
        process_successor(node, next)


# That is a node
# type = struct kmp_base_depnode {
#    kmp_depnode_list_t *successors;
#    kmp_task_t *task;
#    kmp_lock_t lock;
#    volatile kmp_int32 npredecessors;
#    volatile kmp_int32 nrefs;
# }

def process_node(node, address):
    successors, task, npredecessors = values_for(
        address, dn_successors_str, dn_task_str, dn_npredecessors_str)
#    print 'addr 0x%x' % address
    if task:
        routine = get_task_routine(task)
        if not routine:
            return
        current_node = DependentTaskNode(npredecessors, routine)
        node.append_node(current_node)
#        node = current_node
    if successors:
        #        print 'su 0x%x' % successors
        process_successor(node, successors)
    return


def get_depnode_information(dependency, depnode):
    successors, predecessors, task, refs = values_for(depnode,
                                                      dn_successors_str, dn_npredecessors_str, dn_task_str, dn_nrefs_str)
    routine = get_task_routine(task)
    if routine:
        node = DependentTaskNode(predecessors, routine)
        dependency.append_dependent_task(node)
        process_successor(node, successors)


def get_task_and_parents(tasks, task_address):
    if task_address:
        task = get_task_info(task_address)
        if not task.id_ in tasks:
            tasks[task.id_] = task
            parent_address = tasks[task.id_].parent_address
            while parent_address:
                parent_task = get_task_info(parent_address)
                if parent_task:
                    tasks[parent_task.id_] = parent_task
                else:
                    parent_address = 0
                if parent_address != parent_task.parent_address:
                    parent_address = parent_task.parent_address
                else:
                    parent_address = 0


def traverse_task_deque(tasks, found_tasks, deque_address):
    address_size = get_integer_from(address_width_str)
    for i in xrange(found_tasks):
        task_address = read_integer(
            deque_address, address_size, i * address_size)
        get_task_and_parents(tasks, task_address)


def get_task_team(tasks, address):
    threads_data = get_value_for(address, tt_threads_data_str)
    found_tasks = get_value_for(address, tt_found_tasks_str)
    nproc = get_value_for(address, tt_nproc_str)
    unfinished_threads = get_value_for(address, tt_unfinished_threads_str)
    active = get_value_for(address, tt_active_str)
    if found_tasks:
        deque = get_value_for(threads_data, hd_deque_str)
        if not deque:
            return
        traverse_task_deque(tasks, found_tasks, deque)


def get_thread_info(address, gdb_threads):
    th_info = offset_size(th_info_str)
    return ThreadInternal(address + th_info.offset, gdb_threads)


class LockRawInternal(Lock):

    def __init__(self, address, item):
        self.is_valid = True
        initialized, location, tail, \
            head_id, next_ticket, now_serving, \
            owner_id, depth_locked, lock_flags = \
            values_for(address, lk_initialized_str, lk_location_str, lk_tail_id_str,
                       lk_head_id_str, lk_next_ticket_str, lk_now_serving_str,
                       lk_owner_id_str, lk_depth_locked_str, lk_lock_flags_str)

        new_location = get_source_from_ident(
            location) if location else 'undefined'
        self.is_valid = not(
            owner_id is None or depth_locked is None or lock_flags is None)
        status = 'free' if not owner_id else 'held'
        kind = 'nested' if depth_locked >= 0 else 'simple'
        depth_locked = depth_locked if depth_locked >= 0 else 0
        super(LockRawInternal, self).__init__(
            item, new_location, kind, status, depth_locked, owner_id)


class ThreadInternal(Thread):

    def __init__(self, address, gdb_threads):
        self.task_address = 0
        self.team_address = 0
        self.task_team_address_0 = 0
        self.task_team_address_1 = 0
        self.barrier_base = 0
        tid, gtid, handler = values_for(
            address, ds_tid_str, ds_gtid_str, ds_thread_str)
        gdb_id = gdb_threads.get(handler, -1)
        super(ThreadInternal, self).__init__(gtid, handler, tid, gdb_id)

    def __str__(self):
        return '%s 0x%x 0x%x 0x%x 0x%x' % (super(ThreadInternal, self).__str__(),
                                           self.task_address, self.team_address, self.task_team_address_0, self.task_team_address_1)


class TaskFlagBits(ctypes.LittleEndianStructure):
    _fields_ = [
        ('tiedeness', c_uint32, 1),
        ('final', c_uint32, 1),
        ('merged_if0', c_uint32, 1),
        ('destructors_thunk', c_uint32, 1),
        ('reserved', c_uint32, 12),
        ('tasktype', c_uint32, 1),
        ('task_serial', c_uint32, 1),
        ('tasking_ser', c_uint32, 1),
        ('team_serial', c_uint32, 1),
        ('started', c_uint32, 1),
        ('executing', c_uint32, 1),
        ('complete', c_uint32, 1),
        ('freed', c_uint32, 1),
        ('native', c_uint32, 1),
        ('reserved31', c_uint32, 7)
    ]


class TaskFlags(ctypes.Union):
    _fields_ = [
        ('flags', TaskFlagBits),
        ('value', c_uint32)
    ]


def decode_task_flags(flags):
    s = ''
    if flags.executing:
        return 'executing'
    if flags.tiedeness:
        s = 'tied'
    if flags.started:
        s = '%s %s' % (s, 'started')
    return s


class TaskInternal(Task):

    def __init__(self, address):
        id_, flags, team, \
            parent, level, ident, \
            children, incomplete_children, depnode, \
            group = values_for(address, td_task_id_str, td_flags_str, td_team_str,
                               td_parent_str, td_level_str, td_ident_str,
                               td_allocated_child_tasks_str, td_incomplete_child_tasks_str, td_depnode_str,
                               td_taskgroup_str)
        team = get_value_for(team, t_id_str)
        self.parent_address = parent
        parent_id = get_value_for(parent, td_task_id_str)
        flags_decode = TaskFlags()
        flags_decode.value = flags
        str_status = decode_task_flags(flags_decode.flags)
        is_complete = flags_decode.flags.complete
        location = get_ident(ident)
        taskwait_location, taskwait_counter, thread_id = get_taskwait_info(
            address)
        self.taskwait = Taskwait(
            id_ * 100000000 + thread_id, id_, taskwait_location, taskwait_counter, thread_id)
        self.dependency = 0
        if depnode:
            self.dependency = TaskDependency(id_)
            get_depnode_information(self.dependency, depnode)
        super(TaskInternal, self).__init__(id_, flags, team, parent_id, level, location, children,
                                           incomplete_children, str_status, is_complete)


class BarrierInternal(Barrier):

    def __init__(self, id_, address, kind, team):
        master_arrived, workers_arrived = values_for(address,
                                                     t_b_master_arrived_str, t_b_team_arrived_str)
        super(BarrierInternal, self).__init__(
            id_, kind, team, master_arrived, workers_arrived)


class TaskDependencyInternal(TaskDependency):

    def __init__(self, id_, address, kind, team):
        master_arrived, workers_arrived = values_for(address,
                                                     t_b_master_arrived_str,
                                                     t_b_team_arrived_str)
        super(BarrierInternal, self).__init__(
            id_, kind, team, master_arrived, workers_arrived)


class TeamInternal(Team):

    def __init__(self, address):
        team_id, master, parent, \
            nproc, ident, implicit_task, \
            cancel_request = values_for(address, t_id_str, t_master_tid_str, t_parent_str,
                                        t_nproc_str, t_ident_str, t_implicit_task_str,
                                        t_cancel_request_str)
        location = ''
        if ident:
            location = get_ident(ident)
        parent = get_value_for(parent, t_id_str)
        t_bar = offset_size(th_bar_str)
        base_id = 100000000 * team_id + 1
        pointer_size = get_integer_from(address_width_str)
        #pointer_size = t_bar.size
        self.barrier_explicit = BarrierInternal(
            base_id, t_bar.offset + address + 0 * pointer_size, 'implicit', team_id)
        base_id = 100000000 * team_id + 2
        self.barrier_implicit = BarrierInternal(
            base_id, t_bar.offset + address + 1 * pointer_size, 'explicit', team_id)
        base_id = 100000000 * team_id + 3
        self.barrier_reduction = BarrierInternal(
            base_id, t_bar.offset + address + 2 * pointer_size, 'reduction', team_id)
        #print t_bar.size
        #print '0x%x' % (t_bar.offset + address + 0 * pointer_size)
        #print self.barrier_explicit
        #print '0x%x' % (t_bar.offset + address + 1 * pointer_size)
        #print self.barrier_implicit
        #print '0x%x' % (t_bar.offset + address + 2 * 64)
        #print self.barrier_reduction
        super(TeamInternal, self).__init__(
            team_id, master, parent, nproc, location)


def get_barrier(address, task):
    return BarrierInternal(address)


def get_lock_properties(address, item):
    try:
        return LockRawInternal(address, item)
    except Exception as e:
        print e


def traverse_lock_table(table_addr, used, allocate):
    if used > allocate:
        raise OMPMoreLocksThanExpected
    locks = {}
    counter = 0
    pointer_size = get_integer_from(address_width_str)
    while counter < used:
        current_item = table_addr + counter * pointer_size
        item_add = read_integer(current_item, pointer_size)
        current_lock = get_lock_properties(item_add, counter)
        if current_lock is not None and current_lock.is_valid:
            locks[counter] = current_lock
        counter += 1
    return locks


def get_locks():
    lock_table = address_size_padding(lock_table_str)
    used, allocated, table = values_for(
        lock_table.addr, lt_used_str, lt_allocated_str, lt_table_str)
    if table != 0:
        return traverse_lock_table(table, used, allocated)


def traverse_all_threads(threads, gdb_threads, root_or_omp, capacity):
    base_address = address_size_padding(root_or_omp)
    base_address.addr = read_integer(base_address.addr, base_address.size)
    th_info = offset_size(th_info_str)
    th_team = offset_size(th_team_str)
    th_task_team = offset_size(th_task_team_str)
    th_current_task = offset_size(th_current_task_str)
    th_task_team_str_size = get_integer_from(tt_sizeof_struct_str)
    tb_bar = offset_size(th_bar_str)
    thread_iterator = ArrayIterator(
        base_address.addr, base_address.size, capacity)
    current_thread = thread_iterator.advance()
    address_with = get_integer_from(address_width_str)
    while current_thread:
        address = read_integer(
            thread_iterator.current_address, thread_iterator.struct_size, 0)
        if address:
            c_thread = get_thread_info(address + th_info.offset, gdb_threads)
            c_thread.task_team_address_0 = read_integer(
                address + th_task_team.offset, th_task_team.size)
            if c_thread.task_team_address_0:
                c_thread.task_team_address_1 = c_thread.task_team_address_0 + \
                    th_task_team_str_size
            else:
                c_thread.task_team_address_1, c_thread.task_team_address_0 = 0, 0
            c_thread.task_address = read_integer(
                address + th_current_task.offset, th_current_task.size)
            c_thread.team_address = get_value_for(address, th_team_str)
            c_thread.barrier_base = address + tb_bar.offset
            threads[c_thread.handler] = c_thread
        current_thread = thread_iterator.advance()


def get_threads(gdb_threads):
    threads = {}
    capacity = address_size_padding(capacity_debug_str)
    capacity = read_integer(capacity.addr, capacity.size)
    traverse_all_threads(threads, gdb_threads, threads_debug_str, capacity)
    traverse_all_threads(threads, gdb_threads, roots_debug_str, capacity)
    return threads
