# OpenMP support for GDB, the GNU debugger
#
# Copyright (C) 2015 Free Software Foundation, Inc.
#
# Contributed by Intel Corporation, <walfred.tedeschi@intel.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import inspect

import gdb

from openmp_commons import OpenMPRTL, OMPNoInferior
from intel_openmp_dbg import IntelOpenMPRTL


# Implementation of GOMP would start like this.

# class GNUOpenMPRTL(OpenMPRTL):
# Dummy symbol not yet present.
#    debug_symbol = "__ggb"

def get_openmp_interface():
    current_inferior = gdb.selected_inferior()
    if not current_inferior.pid:
        raise OMPNoInferior
    for rtl in openmp_classes:
        if rtl.debug_symbol:
            rtl_instance = rtl(current_inferior)
            if rtl_instance and rtl_instance.has_supported_version():
                return rtl_instance
    raise OMPRTLNotRecognized


def collect_omp_classes():
    return [value for value in globals().itervalues() if inspect.isclass(value) and issubclass(value, OpenMPRTL)]

openmp_classes = collect_omp_classes()
